@@ -26,6 +26,8 @@ Base classes:
 import sys
 import os
 import socket
+from socket import _socket
+from ssl import CertificateError
 import time
 import email
 import codecs
@@ -49,6 +51,35 @@ try:
 except ImportError:
     pass
 
+###
+# Monkey-patch imaplib to support the ID command ([RFC 2971](https://www.rfc-editor.org/rfc/rfc2971)) - begin
+###
+imaplib.Commands['ID'] = ('AUTH', 'SELECTED')
+def imap_id_rfc2971(self, *kv_pairs, **kw):
+    """(typ, [data]) = <instance>.id(kv_pairs)
+    'kv_pairs' is a possibly empty list of keys and values.
+    'data' is a list of ID key value pairs or NIL.
+    NB: a single argument is assumed to be correctly formatted and is passed through unchanged
+    (for backward compatibility with earlier version).
+    Exchange information for problem analysis and determination.
+    The ID extension is defined in RFC 2971. """
+
+    name = 'ID'
+
+    if not kv_pairs:
+        data = 'NIL'
+    elif len(kv_pairs) == 1:
+        data = kv_pairs[0]     # Assume invoker passing correctly formatted string (back-compat)
+    else:
+        data = '(%s)' % ' '.join([(arg and self._quote(arg) or 'NIL') for arg in kv_pairs])
+
+    return self._simple_command(name, data, **kw)
+
+imaplib.IMAP4.id = imap_id_rfc2971
+###
+# Monkey-patch imaplib to support the ID command (RFC 2971) - end
+###
+
 try:
   import ssl
   SSLError = ssl.SSLError
@@ -74,9 +105,11 @@ tocode = lambda x: isinstance(x,bytes) and x or x.encode()
 
 # If we have an ssl module:
 has_sni = ssl and getattr(ssl, 'HAS_SNI', False)
+
 proto_best = ssl and getattr(ssl, 'PROTOCOL_TLS', None)
 if not proto_best:
     proto_best = ssl and getattr(ssl, 'PROTOCOL_SSLv23', None)
+
 has_ciphers = sys.hexversion >= 0x2070000
 
 # Monkey-patch SNI use into SSL.wrap_socket() if supported
@@ -108,104 +141,179 @@ def wrap_socket(sock, keyfile=None, certfile=None,
 if ssl:
     ssl.wrap_socket = wrap_socket
 
-# Is it recent enough to have hostname matching (Python 3.2+)?
-try:
-    ssl_match_hostname = ssl and ssl.match_hostname
-except AttributeError:
-# Running a Python with no hostname matching
-    def _dnsname_match(dn, hostname, max_wildcards=1):
-        """Matching according to RFC 6125, section 6.4.3
-        http://tools.ietf.org/html/rfc6125#section-6.4.3
-        """
-        pats = []
-        if not dn:
-            return False
+##############################################
+# to avoid deprecation warnings
+# match_hostname from Python 3.10 ssl.py
+# starting here down to ssl_match_hostname_END
+##############################################
 
-        parts = dn.split(r'.')
-        leftmost = parts[0]
-        remainder = parts[1:]
+def _dnsname_match(dn, hostname):
+    """Matching according to RFC 6125, section 6.4.3
 
-        wildcards = leftmost.count('*')
-        if wildcards > max_wildcards:
-            # Issue #17980: avoid denials of service by refusing more
-            # than one wildcard per fragment.  A survery of established
-            # policy among SSL implementations showed it to be a
-            # reasonable choice.
-            raise getmailOperationError(
-                "too many wildcards in certificate DNS name: " + repr(dn))
+    - Hostnames are compared lower case.
+    - For IDNA, both dn and hostname must be encoded as IDN A-label (ACE).
+    - Partial wildcards like 'www*.example.org', multiple wildcards, sole
+      wildcard or wildcards in labels other then the left-most label are not
+      supported and a CertificateError is raised.
+    - A wildcard must match at least one character.
+    """
+    if not dn:
+        return False
 
-        # speed up common case w/o wildcards
-        if not wildcards:
-            return dn.lower() == hostname.lower()
+    wildcards = dn.count('*')
+    # speed up common case w/o wildcards
+    if not wildcards:
+        return dn.lower() == hostname.lower()
 
-        # RFC 6125, section 6.4.3, subitem 1.
-        # The client SHOULD NOT attempt to match a presented identifier
-        # in which the wildcard character comprises a label other than
-        # the left-most label.
-        if leftmost == '*':
-            # When '*' is a fragment by itself, it matches a non-empty
-            # dotless fragment.
-            pats.append('[^.]+')
-        elif leftmost.startswith('xn--') or hostname.startswith('xn--'):
-            # RFC 6125, section 6.4.3, subitem 3.
-            # The client SHOULD NOT attempt to match a presented identifier
-            # where the wildcard character is embedded within an A-label or
-            # U-label of an internationalized domain name.
-            pats.append(re.escape(leftmost))
+    if wildcards > 1:
+        raise CertificateError(
+            "too many wildcards in certificate DNS name: {!r}.".format(dn))
+
+    dn_leftmost, sep, dn_remainder = dn.partition('.')
+
+    if '*' in dn_remainder:
+        # Only match wildcard in leftmost segment.
+        raise CertificateError(
+            "wildcard can only be present in the leftmost label: "
+            "{!r}.".format(dn))
+
+    if not sep:
+        # no right side
+        raise CertificateError(
+            "sole wildcard without additional labels are not support: "
+            "{!r}.".format(dn))
+
+    if dn_leftmost != '*':
+        # no partial wildcard matching
+        raise CertificateError(
+            "partial wildcards in leftmost label are not supported: "
+            "{!r}.".format(dn))
+
+    hostname_leftmost, sep, hostname_remainder = hostname.partition('.')
+    if not hostname_leftmost or not sep:
+        # wildcard must match at least one char
+        return False
+    return dn_remainder.lower() == hostname_remainder.lower()
+
+
+def _inet_paton(ipname):
+    """Try to convert an IP address to packed binary form
+
+    Supports IPv4 addresses on all platforms and IPv6 on platforms with IPv6
+    support.
+    """
+    # inet_aton() also accepts strings like '1', '127.1', some also trailing
+    # data like '127.0.0.1 whatever'.
+    try:
+        addr = _socket.inet_aton(ipname)
+    except OSError:
+        # not an IPv4 address
+        pass
+    else:
+        if _socket.inet_ntoa(addr) == ipname:
+            # only accept injective ipnames
+            return addr
         else:
-            # Otherwise, '*' matches any dotless string, e.g. www*
-            pats.append(re.escape(leftmost).replace(r'\*', '[^.]*'))
+            # refuse for short IPv4 notation and additional trailing data
+            raise ValueError(
+                "{!r} is not a quad-dotted IPv4 address.".format(ipname)
+            )
 
-        # add the remaining fragments, ignore any wildcards
-        for frag in remainder:
-            pats.append(re.escape(frag))
+    try:
+        return _socket.inet_pton(_socket.AF_INET6, ipname)
+    except OSError:
+        raise ValueError("{!r} is neither an IPv4 nor an IP6 "
+                         "address.".format(ipname))
+    except AttributeError:
+        # AF_INET6 not available
+        pass
 
-        pat = re.compile(r'\A' + r'\.'.join(pats) + r'\Z', re.IGNORECASE)
-        return pat.match(hostname)
+    raise ValueError("{!r} is not an IPv4 address.".format(ipname))
 
 
-    def ssl_match_hostname(cert, hostname):
-        """Verify that *cert* (in decoded format as returned by
-        SSLSocket.getpeercert()) matches the *hostname*. RFC 2818 and
-        RFC 6125 rules are followed, but IP addresses are not accepted
-        for *hostname*.
+def _ipaddress_match(cert_ipaddress, host_ip):
+    """Exact matching of IP addresses.
 
-        getmailOperationError is raised on failure. On success, the function
-        returns nothing.
-        """
-        if not cert:
-            raise ValueError("empty or no certificate, ssl_match_hostname "
-                             "needs an SSL socket or SSL context with "
-                             "either CERT_OPTIONAL or CERT_REQUIRED")
-        dnsnames = []
-        san = cert.get('subjectAltName', ())
-        for key, value in san:
-            if key == 'DNS':
-                if _dnsname_match(value, hostname):
-                    return
-                dnsnames.append(value)
-        if not dnsnames:
-            # The subject is only checked when there is no dNSName entry
-            # in subjectAltName
-            for sub in cert.get('subject', ()):
-                for key, value in sub:
-                    # XXX according to RFC 2818, the most specific
-                    # Common Name must be used.
-                    if key == 'commonName':
-                        if _dnsname_match(value, hostname):
-                            return
-                        dnsnames.append(value)
-        if len(dnsnames) > 1:
-            raise getmailOperationError("hostname %s "
-                "doesn't match either of %s"
-                % (hostname, ', '.join(map(repr, dnsnames))))
-        elif len(dnsnames) == 1:
-            raise getmailOperationError("hostname %s "
-                "doesn't match %s"
-                % (hostname, dnsnames[0]))
-        else:
-            raise getmailOperationError("no appropriate commonName or "
-                "subjectAltName fields were found")
+    RFC 6125 explicitly doesn't define an algorithm for this
+    (section 1.7.2 - "Out of Scope").
+    """
+    # OpenSSL may add a trailing newline to a subjectAltName's IP address,
+    # commonly with IPv6 addresses. Strip off trailing \n.
+    ip = _inet_paton(cert_ipaddress.rstrip())
+    return ip == host_ip
+
+
+def match_hostname(cert, hostname):
+    """Verify that *cert* (in decoded format as returned by
+    SSLSocket.getpeercert()) matches the *hostname*.  RFC 2818 and RFC 6125
+    rules are followed.
+
+    The function matches IP addresses rather than dNSNames if hostname is a
+    valid ipaddress string. IPv4 addresses are supported on all platforms.
+    IPv6 addresses are supported on platforms with IPv6 support (AF_INET6
+    and inet_pton).
+
+    CertificateError is raised on failure. On success, the function
+    returns nothing.
+    """
+    # warnings.warn(
+    #     "ssl.match_hostname() is deprecated",
+    #     category=DeprecationWarning,
+    #     stacklevel=2
+    # )
+    if not cert:
+        raise ValueError("empty or no certificate, match_hostname needs a "
+                         "SSL socket or SSL context with either "
+                         "CERT_OPTIONAL or CERT_REQUIRED")
+    try:
+        host_ip = _inet_paton(hostname)
+    except ValueError:
+        # Not an IP address (common case)
+        host_ip = None
+    dnsnames = []
+    san = cert.get('subjectAltName', ())
+    for key, value in san:
+        if key == 'DNS':
+            if host_ip is None and _dnsname_match(value, hostname):
+                return
+            dnsnames.append(value)
+        elif key == 'IP Address':
+            if host_ip is not None and _ipaddress_match(value, host_ip):
+                return
+            dnsnames.append(value)
+    if not dnsnames:
+        # The subject is only checked when there is no dNSName entry
+        # in subjectAltName
+        for sub in cert.get('subject', ()):
+            for key, value in sub:
+                # XXX according to RFC 2818, the most specific Common Name
+                # must be used.
+                if key == 'commonName':
+                    if _dnsname_match(value, hostname):
+                        return
+                    dnsnames.append(value)
+    if len(dnsnames) > 1:
+        raise CertificateError("hostname %r "
+            "doesn't match either of %s"
+            % (hostname, ', '.join(map(repr, dnsnames))))
+    elif len(dnsnames) == 1:
+        raise CertificateError("hostname %r "
+            "doesn't match %r"
+            % (hostname, dnsnames[0]))
+    else:
+        raise CertificateError("no appropriate commonName or "
+            "subjectAltName fields were found")
+
+def ssl_match_hostname(cert, hostname):
+    try:
+        return match_hostname(cert, hostname)
+    except (ValueError, CertificateError) as e:
+        raise getmailOperationError("%s"%e)
+
+##############################################
+# match_hostname from Python 3.10 ssl.py
+# ssl_match_hostname_END
+##############################################
 
 from getmailcore.exceptions import *
 from getmailcore.constants import *
@@ -524,7 +632,13 @@ class IMAP4_SSL_EXTENDED(imaplib.IMAP4_SSL):
        self.ssl_version = ssl_version
        self.ca_certs = ca_certs
        self.ssl_ciphers = ssl_ciphers
-       imaplib.IMAP4_SSL.__init__(self, host, port, keyfile, certfile)
+       self.keyfile = keyfile
+       self.certfile = certfile
+       try:
+           imaplib.IMAP4_SSL.__init__(self, host, port, keyfile, certfile)
+       except TypeError:
+           imaplib.IMAP4_SSL.__init__(self, host, port)
+
 
     def open(self, host='', port=imaplib.IMAP4_SSL_PORT, timeout=None):
        self.host = host
@@ -669,7 +783,7 @@ class RetrieverSkeleton(ConfigurableBase):
 
       __str__(self) - return a simple string representing the class instance.
 
-      _getmsglist(self) - retieve a list of all available messages, and store
+      _getmsglist(self) - retrieve a list of all available messages, and store
                           unique message identifiers in the dict
                           self.msgnum_by_msgid.
                           Message identifiers must be unique and persistent
@@ -677,10 +791,11 @@ class RetrieverSkeleton(ConfigurableBase):
                           octets) in a dictionary self.msgsizes, using the
                           message identifiers as keys.
 
-      _delmsgbyid(self, msgid) - delete a message from the message store based
-                                 on its message identifier.
+      _flagmsgbyid(self, msgid) - On default flag message msgid for deletion.
+                                  For IMAP imap_on_delete allows other flag:
+                                  return True if `Deleted` is among the flags.
 
-      _getmsgbyid(self, msgid) - retreive and return a message from the message
+      _getmsgbyid(self, msgid) - retrieve and return a message from the message
                                  store based on its message identifier.  The
                                  message is returned as a Message() class
                                  object. The message will have additional data
@@ -716,6 +831,7 @@ class RetrieverSkeleton(ConfigurableBase):
         self._clear_state()
         self.conn = None
         self.supports_idle = False
+        self.supports_id = False
         ConfigurableBase.__init__(self, **args)
 
     def set_new_timestamp(self):
@@ -899,6 +1015,9 @@ class RetrieverSkeleton(ConfigurableBase):
 
     def delivered(self, msgid):
         self.__delivered[msgid] = None
+        if self.app_options.get('to_oldmail_on_each_mail',False):
+            self.write_oldmailfile(self.mailbox_selected)
+            self.__delivered = {}
 
     def getheader(self, msgid):
         if not self.__initialized:
@@ -923,8 +1042,9 @@ class RetrieverSkeleton(ConfigurableBase):
     def delmsg(self, msgid):
         if not self.__initialized:
             raise getmailOperationError('not initialized')
-        self._delmsgbyid(msgid)
-        self.deleted[msgid] = True
+        deleted = self._flagmsgbyid(msgid)
+        self.deleted[msgid] = deleted
+        return deleted
 
     def run_password_command(self):
         command = self.conf['password_command'][0]
@@ -932,12 +1052,12 @@ class RetrieverSkeleton(ConfigurableBase):
         rc, stdout, stderr = run_command(command,args)
         if rc:
             raise getmailOperationError(
-                'External program error (%s exited with %d)' % (args[0],rc)
+                'External program error (%s exited with %d)' % (command,rc)
             )
         if stderr:
             self.log.warning(
                 'External password program "%s" wrote to stderr: %s'
-                % (args[0], stderr)
+                % (command, stderr)
             )
         password = stdout
         self.conf['password'] = password
@@ -1049,10 +1169,11 @@ class POP3RetrieverBase(RetrieverSkeleton):
             )
         self.gotmsglist = True
 
-    def _delmsgbyid(self, msgid):
+    def _flagmsgbyid(self, msgid):
         self.log.trace()
         msgnum = self._getmsgnumbyid(msgid)
         self.conn.dele(msgnum)
+        return True
 
     def _getmsgbyid(self, msgid):
         self.log.debug('msgid %s' % msgid + os.linesep)
@@ -1343,6 +1464,10 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         self.log.trace('got %s' % r + os.linesep)
         return r
 
+    def id(self):
+        server_id = self.conn.id('name', 'getmail', 'version', '6.0.0')
+        return server_id
+
     def list_mailboxes(self):
         '''List (selectable) IMAP folders in account.'''
         cmd = ('LIST', )
@@ -1355,7 +1480,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         # Close current mailbox so deleted mail is expunged.  One getmail
         # user had a buggy IMAP server that didn't do the automatic expunge,
         # so we do it explicitly here if we've deleted any messages.
-        if self.deleted:
+        if any(self.deleted):
             self.conn.expunge()
         self.write_oldmailfile(self.mailbox_selected)
         # And clear some state
@@ -1413,7 +1538,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             self.mailbox_selected = mailbox
             # use *last* EXISTS returned
             count = int(count[-1])
-            uidvalidity = tostr(self.conn.response('UIDVALIDITY')[1][0])
+            uidvalidity = tostr(self.conn.response('UIDVALIDITY')[1][0] or b"") or None
         except imaplib.IMAP4.error as o:
             raise getmailOperationError('IMAP error (%s)' % o)
         except (IndexError, ValueError) as o:
@@ -1497,7 +1622,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                     self._mboxuids[msgid] = r['uid']
                     self._mboxuidorder.append(msgid)
                     self.msgnum_by_msgid[msgid] = None
-                    self.msgsizes[msgid] = (0 if 
+                    self.msgsizes[msgid] = (0 if
                         self.app_options['skip_imap_fetch_size']
                                         else int(r['rfc822.size']))
 
@@ -1520,7 +1645,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
     def __getitem__(self, i):
         return self._mboxuidorder[i]
 
-    def _delmsgbyid(self, msgid):
+    def _flagmsgbyid(self, msgid):
         self.log.trace()
         try:
             uid = self._getmboxuidbymsgid(msgid)
@@ -1533,10 +1658,16 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                     'COPY', uid, self.conf['move_on_delete']
                 )
             self.log.debug('deleting message "%s"' % uid + os.linesep)
+
+            # On GMail you need to remove LABELS before deleting
+            if 'X-GM-EXT-1' in self.conn.capabilities:
+                self._parse_imapuidcmdresponse('STORE', uid,'X-GM-LABELS', '()')
+
+            flag = self.conf.get('imap_on_delete',None) or r'(\Deleted \Seen)'
             response = self._parse_imapuidcmdresponse(
-                'STORE', uid, 'FLAGS',
-                self.conf.get('imap_on_delete',None) or r'(\Deleted \Seen)'
+                'STORE', uid, 'FLAGS', flag
             )
+            return 'Deleted' in flag
         except imaplib.IMAP4.error as o:
             raise getmailOperationError('IMAP error (%s)' % o)
 
@@ -1634,7 +1765,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         results = ext.groupdict()
         metadata = {}
         for item in ('LABELS', 'THRID', 'MSGID'):
-            if item in results and results[item]:
+            if results.get(item):
                 metadata['X-GMAIL-%s' % item] = results[item]
 
         return metadata
@@ -1734,6 +1865,12 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                 self.supports_idle = True
                 imaplib.Commands['IDLE'] = ('AUTH', 'SELECTED')
 
+            if 'ID' in self.conn.capabilities:
+                self.supports_id = True
+
+            if self.supports_id and self.conf['imap_id_extension']:
+                self.id()
+
             if self.mailboxes == ('ALL', ):
                 # Special value meaning all mailboxes in account
                 self.mailboxes = tuple(self.list_mailboxes())
